//////////////////////////////////////////////////////////////////////////
///	load PCM wave files, convert to 44100 K 16 bit mono, put into ////
///	loopData[] and loopLength[], handle mutex locks for each loop ////
//////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include "loopdata.h"

extern pthread_mutex_t loopMutex[];

void lockLoop(int L)
{
        pthread_mutex_lock(&loopMutex[L]);
}

void unlockLoop(int L)
{
        pthread_mutex_unlock(&loopMutex[L]);
}

int isLoopLocked(int L)
{
	int ret;
        ret = pthread_mutex_trylock(&loopMutex[L]);
        if (ret == 0) pthread_mutex_unlock(&loopMutex[L]);
        return (ret != 0);
}

void destroyLoopMutexes()
{
	int i;
	for (i=0;i<NUM_LOOPS;i++)
		pthread_mutex_destroy(&loopMutex[i]);
}

void deleteLoop(int L)
{
	if (loopData[L]!=NULL)
		free (loopData[L]);
	loopLength[L]=0;
	loopPos[L]=0;
	loopData[L] = NULL;
}

int isLoopFull(int L)  {return (loopData[L]!=NULL);}

static inline short conv8 (char c)
{
        short ret = c; ret -= 0x80; ret <<= 8; return ret;
}

static inline short av16 (short a, short b)
{
        int sum = a + b; sum /=2; return (short) sum;
}

static inline short GetNext16Mono (char ** raw, int Channels, int Bits)
{
        short a[2];
        int i = 0;
        for (i=0;i<Channels;i++)
        {
                switch (Bits)
                {
                        case 8:
                                a[i] = conv8( **raw);
                                (*raw)++;
                                break;
                        case 16:
                                a[i] = * ( (short *) *raw);
                                (*raw)+=2;
                                break;
                };
        }
        if (Channels == 1) return a[0];
        return av16( a[0], a[1] );

}

char llError[1024];

int loadLoop(int L, char * fileName)
{
	int fd;
	int fileLength;
	FILE * wavFile;
	char wavHdr[512];
	unsigned int hdrLen;
	int fmtOffset = -1;
	int i;
	int Bits, HZ, Channels;
	short format;
	int dataOffset = -1;
	int rawSize;
	char * rawData;
	int loopSize;
	short * pLoop;
	char * raw;
	char * endraw;
	short Get; int padHZ;


	
	deleteLoop(L);
	llError[0]=0;
	// now we have to figure out the length of the file, open it,
	// get the header to figure out format and data offset,
	// allocate proper amount of space, and convert file's WAV format
	// to 16-bit 44.1K mono
	
	// to get file length, we must open and close it first with low-level routines
	// so that we can use lseek instead of fseek (which doesn't give length)
	
	fd = open(fileName, O_RDONLY);
	if (fd == -1)
	{
		sprintf(llError, "Error opening file: %s", fileName);
		return 0;
	}
	fileLength = lseek(fd, 0, SEEK_END);
	if (fileLength == -1)
	{
		sprintf(llError, "Error determining file length: %s", fileName);
		return 0;
	}
	close(fd);
	
	// open the file with standard I/O library
	wavFile = fopen(fileName, "rb");
	if (wavFile == NULL)
	{
		sprintf(llError, "Error opening file: %s", fileName);
		return 0;
	}
	memset (wavHdr, 0, 512);
	hdrLen = (fileLength < 512) ? fileLength : 512;
	if (fread( wavHdr, 1, hdrLen, wavFile) != hdrLen)
	{
		sprintf(llError, "Error reading header from file: %s", fileName);
		fclose(wavFile);
		return 0;
	}
	// is it a WAVE file?
	if ( (strncmp (&wavHdr[0], "RIFF", 4) != 0) || (strncmp (&wavHdr[8], "WAVE", 4) != 0) )
	{
		sprintf(llError, "Not a wave file: %s", fileName);
		fclose(wavFile);
		return 0;
	}
	
	// now search for "fmt " tag
	
	for (i=0;i<508;i++)
	{
		if (strncmp (&wavHdr[i], "fmt ", 4) == 0)
		{
			fmtOffset = i; break;
		}
	}
	if (fmtOffset == -1)
	{
		sprintf(llError, "fmt tag not found in: %s", fileName);
		fclose(wavFile);
		return 0;
	}
	
	// read format out of file

	format = * ( (short *) &wavHdr[fmtOffset + 8] );
	if (format != 1)
	{
		sprintf(llError, "Not a PCM wave file: %s", fileName);
		fclose(wavFile);
		return 0;
	}
	Channels = * ( (short *) &wavHdr[fmtOffset + 10] );
	HZ = * ( (int *) &wavHdr[fmtOffset + 12] );
	Bits = * ( (short *) &wavHdr[fmtOffset + 22] );
	
	switch (HZ)
	{
		case 11025:
		case 22050:
		case 44100:
			break;
		default:
			sprintf(llError, "Illegal HZ %d: %s", HZ, fileName);
		fclose(wavFile);
		return 0;
	};
	
	switch (Channels)
	{
		case 1:
		case 2:
			break;
		default:
			sprintf(llError, "Illegal Channels %d: %s", Channels, fileName);
		fclose(wavFile);
		return 0;
	};
	
	switch (Bits)
	{
		case 8:
		case 16:
			break;
		default:
			sprintf(llError, "Illegal Bits %d: %s", Bits, fileName);
		fclose(wavFile);
		return 0;
	};
	
	

	// now search for "data" tag
	for (;i<508;i++)
	{
		if (strncmp (&wavHdr[i], "data", 4) == 0)
		{
			dataOffset = i; break;
		}
		
	}
	
	if (dataOffset == -1)
	{
		sprintf(llError, "data tag not found in: %s", fileName);
		fclose(wavFile);
		return 0;
	}
	
	dataOffset += 8;
	if (dataOffset >= fileLength)
	{
		sprintf(llError, "File is missing data or corrupted: %s", fileName);
		fclose(wavFile);
		return 0;
	}
	// now read in the data in the original format
	rawSize = fileLength - dataOffset;
	rawData = malloc(rawSize);
	fseek(wavFile, dataOffset, SEEK_SET);
	if (fread (rawData, 1, (size_t) rawSize, wavFile) != (size_t)rawSize)
	{
		sprintf(llError, "Error reading data from file: %s", fileName);
		free(rawData);
		fclose(wavFile);
		return 0;
	}
	
	loopSize = rawSize;
	loopSize /= Channels; // stereo is averaged to mono
	loopSize *= (44100 / HZ); // lower rates are padded
	loopSize *= 16/Bits; // 8-bit samples are converted to 16
	
	loopData[L] = malloc(loopSize);
	pLoop = (short *) loopData[L];
	raw = rawData;
	endraw = rawData + rawSize;
	
	padHZ = 44100/ HZ;
	while (raw < endraw)
	{
		Get = GetNext16Mono(&raw, Channels, Bits);
		for (i=0;i<padHZ;i++)
		{
			*pLoop = Get;
			pLoop++;
		}
		
	}
	
	loopLength[L] = loopSize;
	free(rawData);
	fclose(wavFile);
	return 1;
}
